package answer

import (
	"encoding/json"
	"log/slog"
	"net/url"
	"strings"

	"gitea.stevedudenhoeffer.com/steve/answer/pkg/search"

	gollm "gitea.stevedudenhoeffer.com/steve/go-llm"
)

type searchResults struct {
	Url    string `json:"url"`
	Answer string `json:"answer"`
}

func (s searchResults) String() (string, error) {
	b, err := json.Marshal(s)
	if err != nil {
		return "", err
	}

	return string(b), nil
}

func pickResult(ctx *gollm.Context, results []search.Result, q Question) (*search.Result, error) {
	// if there's only one result, return it
	if len(results) == 1 {
		return &results[0], nil
	}

	// if there are no results, return nil
	if len(results) == 0 {
		return nil, nil
	}

	var pick *search.Result
	var refused bool
	// finally, if there are multiple results then ask the LLM to pick one to read next
	fnPick := gollm.NewFunction(
		"pick",
		"The search result to read next.",
		func(ctx *gollm.Context, args struct {
			URL string `description:"the url to read next"`
		}) (string, error) {
			for _, r := range results {
				if r.URL == args.URL {
					pick = &r
					break
				}
			}

			return "", nil
		})

	fnNoPick := gollm.NewFunction(
		"no_pick",
		"Indicate that there are no results worth reading.",
		func(ctx *gollm.Context, args struct {
			Ignored string `description:"ignored, just here to make sure the function is called.  Fill with anything."`
		}) (string, error) {
			refused = true
			return "", nil
		})

	req := gollm.Request{
		Messages: []gollm.Message{
			{
				Role: gollm.RoleSystem,
				Text: `You are being given results from a web search.  Please select the result you would like to read next to answer the question.  Try to pick the most reputable and relevant result.
The results will be in the JSON format of: {"Url": "https://url.here", "Title": "Title Of Search", "Description": "description here"}`,
			},
			{
				Role: gollm.RoleSystem,
				Text: "The question you are trying to answer is: " + q.Question,
			},
		},
		Toolbox: gollm.NewToolBox(fnPick, fnNoPick),
	}

	for _, r := range results {
		b, _ := json.Marshal(r)
		req.Messages = append(req.Messages, gollm.Message{
			Role: gollm.RoleUser,
			Text: string(b),
		})
	}

	res, err := q.Model.ChatComplete(ctx, req)

	if err != nil {
		return nil, err
	}

	if len(res.Choices) == 0 {
		return nil, nil
	}

	if len(res.Choices[0].Calls) == 0 {
		return nil, nil
	}

	_, _ = req.Toolbox.Execute(ctx, res.Choices[0].Calls[0])

	if refused {
		return nil, nil
	}

	return pick, nil
}

func internalSearch(ctx *gollm.Context, q Question, searchTerm string) (searchResults, error) {
	slog.Info("searching", "search", searchTerm, "question", q)
	results, err := q.Search.Search(ctx, searchTerm)
	if err != nil {
		return searchResults{}, err
	}

	if len(results) == 0 {
		return searchResults{Url: "not-found", Answer: "no search results found"}, nil
	}

	for len(results) > 0 {
		var pick *search.Result
		if len(results) == 1 {
			pick = &results[0]
			results = results[1:]
		} else {
			var err error
			pick, err = pickResult(ctx, results, q)

			slog.Info("picked result", "result", pick, "error", err)
			if err != nil {
				return searchResults{}, err
			}

			if pick == nil {
				break
			}
		}

		trimmed := strings.TrimSpace(pick.URL)
		if trimmed == "" {

		}

		slog.Info("extracting article", "url", trimmed)
		u, err := url.Parse(trimmed)
		if err != nil {
			continue
		}

		a, err := extractArticle(ctx, q.Cache, u)

		if err != nil {
			continue
		}

		slog.Info("extracted article", "url", a.URL, "title", a.Title, "body", a.Body)

		if a.Title != "" && a.Body != "" {
			answer, err := doesTextAnswerQuestion(ctx, q, a.Body)

			if err != nil {
				slog.Error("error checking if text answers question", "question", q.Question, "error", err)
				continue
			}

			if answer != "" {
				return searchResults{Url: u.String(), Answer: answer}, nil
			}
		}
	}

	return searchResults{Url: "not-found", Answer: "no searched results answered"}, nil
}

func functionSearch(ctx *gollm.Context, q Question, searchTerm string) (string, error) {
	res, err := internalSearch(ctx, q, searchTerm)
	if err != nil {
		return "", err
	}

	return res.String()
}
